"""AUTOGENERATED. DO NOT EDIT."""

from keras.__internal__ import backend
from keras.__internal__ import layers
from keras.__internal__ import losses
from keras.__internal__ import models
from keras.__internal__ import optimizers
from keras.__internal__ import utils
from keras.src.engine.base_layer import _apply_name_scope_on_model_declaration as apply_name_scope_on_model_declaration
from keras.src.engine.keras_tensor import KerasTensor
from keras.src.engine.keras_tensor import RaggedKerasTensor
from keras.src.engine.keras_tensor import SparseKerasTensor
from keras.src.saving.serialization_lib import enable_unsafe_deserialization
