"""AUTOGENERATED. DO NOT EDIT."""

from keras.src.initializers import deserialize
from keras.src.initializers import get
from keras.src.initializers import serialize
from keras.src.initializers.initializers import Constant
from keras.src.initializers.initializers import Constant as constant
from keras.src.initializers.initializers import GlorotNormal
from keras.src.initializers.initializers import GlorotNormal as glorot_normal
from keras.src.initializers.initializers import GlorotUniform
from keras.src.initializers.initializers import GlorotUniform as glorot_uniform
from keras.src.initializers.initializers import HeNormal
from keras.src.initializers.initializers import HeNormal as he_normal
from keras.src.initializers.initializers import HeUniform
from keras.src.initializers.initializers import HeUniform as he_uniform
from keras.src.initializers.initializers import Identity
from keras.src.initializers.initializers import Identity as identity
from keras.src.initializers.initializers import Initializer
from keras.src.initializers.initializers import LecunNormal
from keras.src.initializers.initializers import LecunNormal as lecun_normal
from keras.src.initializers.initializers import LecunUniform
from keras.src.initializers.initializers import LecunUniform as lecun_uniform
from keras.src.initializers.initializers import Ones
from keras.src.initializers.initializers import Ones as ones
from keras.src.initializers.initializers import Orthogonal
from keras.src.initializers.initializers import Orthogonal as orthogonal
from keras.src.initializers.initializers import RandomNormal
from keras.src.initializers.initializers import RandomNormal as random_normal
from keras.src.initializers.initializers import RandomUniform
from keras.src.initializers.initializers import RandomUniform as random_uniform
from keras.src.initializers.initializers import TruncatedNormal
from keras.src.initializers.initializers import TruncatedNormal as truncated_normal
from keras.src.initializers.initializers import VarianceScaling
from keras.src.initializers.initializers import VarianceScaling as variance_scaling
from keras.src.initializers.initializers import Zeros
from keras.src.initializers.initializers import Zeros as zeros
