"""AUTOGENERATED. DO NOT EDIT."""

from keras.src.mixed_precision.loss_scale_optimizer import BaseLossScaleOptimizer as LossScaleOptimizer
from keras.src.mixed_precision.policy import Policy
from keras.src.mixed_precision.policy import global_policy
from keras.src.mixed_precision.policy import set_global_policy
